/*	$OpenBSD: pf_ioctl.c,v 1.8 2002/08/12 16:41:25 dhartmei Exp $ */

/*
 * Copyright (c) 2001 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Effort sponsored in part by the Defense Advanced Research Projects
 * Agency (DARPA) and Air Force Research Laboratory, Air Force
 * Materiel Command, USAF, under agreement number F30602-01-2-0537.
 *
 */
#if defined(__FreeBSD__)
#include <sys/types.h>
#endif
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/filio.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/kernel.h>
#include <sys/time.h>
#include <sys/conf.h>
#if defined(__FreeBSD__)
#include <sys/queue.h>
#include <vm/uma.h>
#else
#include <sys/timeout.h>
#include <sys/pool.h>
#endif

#include <net/if.h>
#include <net/if_types.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>

#if !defined(__FreeBSD__)
#include <net/pfvar.h>
#endif

#ifdef INET6
#include <netinet/ip6.h>
#include <netinet/in_pcb.h>
#endif /* INET6 */

#if defined(__FreeBSD__)
#include <machine/limits.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/protosw.h>
#if defined(INET6)
#include <netinet6/ip6protosw.h>
#endif
#include <net/pfil.h>
#include "pfvar.h"
#include "pf_support.h"
#endif

#if defined(__FreeBSD__)
void    init_zone_var(void);
void    cleanup_pf_zone(void);
int	pfattach(void);
#else
void	pfattach(int);
#endif
#if defined(__FreeBSD__)
int 	pfopen(dev_t dev, int flags, int fmt, struct thread *td);
int	pfclose(dev_t dev, int flags, int fmt, struct thread *td);
int	pfioctl(dev_t dev, u_long cmd, caddr_t addr, int flags, struct thread *td);

extern struct callout_handle	pf_expire_to;
/*
 * actually, this should go <sys/protosw.h>
 */
extern struct protosw inetsw[];
#else
int		pfopen(dev_t, int, int, struct proc *);
int		pfclose(dev_t, int, int, struct proc *);
int		pfioctl(dev_t, u_long, caddr_t, int, struct proc *);

extern struct timeout		pf_expire_to;
#endif


#define DPFPRINTF(n, x) if (pf_status.debug >= (n)) printf x

#if defined(__FreeBSD__)
static dev_t		pf_dev;
#define PF_NAME		"pf"
#define PF_CDEV_MAJOR	(73)

static int pf_beginrules(void *addr);
static int pf_commitrules(void *addr);
static int pf_beginnats(void *addr);
static int pf_commitnats(void *addr);
static int pf_beginbinats(void *addr);
static int pf_commitbinats(void *addr);
static int pf_beginrdrs(void *addr);
static int pf_commitrdrs(void *addr);
static int pf_clrstates(void);
static int pf_check(void *ip, int hlen, struct ifnet *ifp, int dir,
		struct mbuf **m);
#if defined(INET6)
static int pf_check6(void *ip, int hlen, struct ifnet *ifp, int dir,
		struct mbuf **m);
#endif
static void hook_pf(void);
static void dehook_pf(void);

static struct cdevsw pf_cdevsw = {
	/* open */	pfopen,
	/* close */	pfclose,
	/* read */	noread,
	/* write */	nowrite,
	/* ioctl */	pfioctl,
	/* poll */	nopoll,
	/* mmap */	nommap,
	/* strategy */	nostrategy,
	/* name */	"pf",
	/* maj */	PF_CDEV_MAJOR,
	/* dump */	nodump,
	/* psize */	nopsize,
	/* flags */	0,
	/* kqfilter */	nokqfilter,
};
#endif

#if defined(__FreeBSD__)
void
init_zone_var(void)
{
	pf_tree_pl = pf_rule_pl = pf_nat_pl = pf_binat_pl = NULL;
	pf_rdr_pl = pf_state_pl = pf_addr_pl = NULL;
	pf_frent_pl = pf_frag_pl = pf_cache_pl = pf_cent_pl = NULL;
}

void
cleanup_pf_zone(void)
{
	if(pf_tree_pl)
		uma_zdestroy(pf_tree_pl);
	if(pf_rule_pl)
		uma_zdestroy(pf_rule_pl);
	if(pf_nat_pl)
		uma_zdestroy(pf_nat_pl);
	if(pf_binat_pl)
		uma_zdestroy(pf_binat_pl);
	if(pf_rdr_pl)
		uma_zdestroy(pf_rdr_pl);
	if(pf_state_pl)
		uma_zdestroy(pf_state_pl);
	if(pf_addr_pl)
		uma_zdestroy(pf_addr_pl);
	if(pf_frent_pl)
		uma_zdestroy(pf_frent_pl);
	if(pf_frag_pl)
		uma_zdestroy(pf_frag_pl);
	if(pf_cache_pl)
		uma_zdestroy(pf_cache_pl);
	if(pf_cent_pl)
		uma_zdestroy(pf_cent_pl);
}
#endif

#if defined(__FreeBSD__)
int
pfattach(void)
#else
void
pfattach(int num)
#endif
{
#if defined(__FreeBSD__)
	pf_tree_pl = uma_zcreate("pftrpl", sizeof(struct pf_tree_node),
		NULL, NULL, NULL, NULL, UMA_ALIGN_PTR, 0);
	if (pf_tree_pl == NULL) {
		cleanup_pf_zone();
		return (1);
	}
	pf_rule_pl = uma_zcreate("pfrulepl", sizeof(struct pf_rule),
		NULL, NULL, NULL, NULL, UMA_ALIGN_PTR, 0);
	if (pf_rule_pl == NULL) {
		cleanup_pf_zone();
		return (1);
	}
	pf_nat_pl = uma_zcreate("pfnatpl", sizeof(struct pf_nat),
		NULL, NULL, NULL, NULL, UMA_ALIGN_PTR, 0);
	if (pf_nat_pl == NULL) {
		cleanup_pf_zone();
		return (1);
	}
	pf_binat_pl = uma_zcreate("pfbinatpl", sizeof(struct pf_binat),
		NULL, NULL, NULL, NULL, UMA_ALIGN_PTR, 0);
	if (pf_binat_pl == NULL) {
		cleanup_pf_zone();
		return (1);
	}
	pf_rdr_pl = uma_zcreate("pfrdrpl", sizeof(struct pf_rdr),
		NULL, NULL, NULL, NULL, UMA_ALIGN_PTR, 0);
	if (pf_rdr_pl == NULL) {
		cleanup_pf_zone();
		return (1);
	}
	pf_state_pl = uma_zcreate("pfstatepl", sizeof(struct pf_state),
		NULL, NULL, NULL, NULL, UMA_ALIGN_PTR, 0);
	if (pf_state_pl == NULL) {
		cleanup_pf_zone();
		return (1);
	}
	pf_addr_pl = uma_zcreate("pfaddr", sizeof(struct pf_addr_dyn),
		NULL, NULL, NULL, NULL, UMA_ALIGN_PTR, 0);
	if (pf_addr_pl == NULL) {
		cleanup_pf_zone();
		return (1);
	}
#else
	 pool_init(&pf_tree_pl, sizeof(struct pf_tree_node), 0, 0, 0, "pftrpl",
	     NULL);
	 pool_init(&pf_rule_pl, sizeof(struct pf_rule), 0, 0, 0, "pfrulepl",
	     &pool_allocator_nointr);
	 pool_init(&pf_nat_pl, sizeof(struct pf_nat), 0, 0, 0, "pfnatpl",
	     &pool_allocator_nointr);
	 pool_init(&pf_binat_pl, sizeof(struct pf_binat), 0, 0, 0, "pfbinatpl",
	     &pool_allocator_nointr);
	 pool_init(&pf_rdr_pl, sizeof(struct pf_rdr), 0, 0, 0, "pfrdrpl",
	     &pool_allocator_nointr);
	 pool_init(&pf_state_pl, sizeof(struct pf_state), 0, 0, 0, "pfstatepl",
	     NULL);
	 pool_init(&pf_addr_pl, sizeof(struct pf_addr_dyn), 0, 0, 0, "pfaddr",
	     NULL);
#endif

	TAILQ_INIT(&pf_rules[0]);
	TAILQ_INIT(&pf_rules[1]);
	TAILQ_INIT(&pf_nats[0]);
	TAILQ_INIT(&pf_nats[1]);
	TAILQ_INIT(&pf_binats[0]);
	TAILQ_INIT(&pf_binats[1]);
	TAILQ_INIT(&pf_rdrs[0]);
	TAILQ_INIT(&pf_rdrs[1]);
	pf_rules_active = &pf_rules[0];
	pf_rules_inactive = &pf_rules[1];
	pf_nats_active = &pf_nats[0];
	pf_nats_inactive = &pf_nats[1];
	pf_binats_active = &pf_binats[0];
	pf_binats_inactive = &pf_binats[1];
	pf_rdrs_active = &pf_rdrs[0];
	pf_rdrs_inactive = &pf_rdrs[1];

#if defined(__FreeBSD__)
	pf_expire_to = timeout(pf_purge_timeout, NULL, pftm_interval * hz);
#else
	timeout_set(&pf_expire_to, pf_purge_timeout, &pf_expire_to);
	timeout_add(&pf_expire_to, pftm_interval * hz);
#endif

#if defined(__FreeBSD__)
	if (pf_normalize_init()) {
		cleanup_pf_zone();
		return (1);
	}
	pf_pool_limits[0].pp = pf_state_pl;
	pf_pool_limits[0].limit = UINT_MAX;
	pf_pool_limits[1].pp = pf_frent_pl;
	pf_pool_limits[1].limit = PFFRAG_FRENT_HIWAT;
#else
	pf_normalize_init();
#endif
	pf_status.debug = PF_DEBUG_URGENT;
#if defined(__FreeBSD__)
	return (0);
#endif
}

int
#if defined(__FreeBSD__)
pfopen(dev_t dev, int flags, int fmt, struct thread *td)
#else
pfopen(dev_t dev, int flags, int fmt, struct proc *p)
#endif
{
#if 0
#if defined(__FreeBSD__)
	int error;

	error = suser(td);
	if (error)
		return (error);
#endif
#endif
	if (minor(dev) >= 1)
		return (ENXIO);
	return (0);
}

int
#if defined(__FreeBSD__)
pfclose(dev_t dev, int flags, int fmt, struct thread *td)
#else
pfclose(dev_t dev, int flags, int fmt, struct proc *p)
#endif
{
	if (minor(dev) >= 1)
		return (ENXIO);
	return (0);
}

int
#if defined(__FreeBSD__)
pfioctl(dev_t dev, u_long cmd, caddr_t addr, int flags, struct thread *td)
#else
pfioctl(dev_t dev, u_long cmd, caddr_t addr, int flags, struct proc *p)
#endif
{
	int error = 0;
	int s;

	/* XXX keep in sync with switch() below */
	if (securelevel > 1)
		switch (cmd) {
		case DIOCGETRULES:
		case DIOCGETRULE:
		case DIOCGETNATS:
		case DIOCGETNAT:
		case DIOCGETBINATS:
		case DIOCGETBINAT:
		case DIOCGETRDRS:
		case DIOCGETRDR:
		case DIOCGETSTATE:
		case DIOCSETSTATUSIF:
		case DIOCGETSTATUS:
		case DIOCCLRSTATUS:
		case DIOCNATLOOK:
		case DIOCSETDEBUG:
		case DIOCGETSTATES:
		case DIOCGETTIMEOUT:
		case DIOCCLRRULECTRS:
		case DIOCGETLIMIT:
			break;
		default:
			return (EPERM);
		}

	if (!(flags & FWRITE))
		switch (cmd) {
		case DIOCGETRULES:
		case DIOCGETRULE:
		case DIOCGETNATS:
		case DIOCGETNAT:
		case DIOCGETRDRS:
		case DIOCGETRDR:
		case DIOCGETSTATE:
		case DIOCGETSTATUS:
		case DIOCGETSTATES:
		case DIOCGETTIMEOUT:
		case DIOCGETBINATS:
		case DIOCGETBINAT:
		case DIOCGETLIMIT:
			break;
		default:
			return (EACCES);
		}

#if defined(__FreeBSD__)
	mtx_lock(&pf_rule_mtx);
#endif

	switch (cmd) {

	case DIOCSTART:
		if (pf_status.running)
			error = EEXIST;
		else {
			u_int32_t states = pf_status.states;
			bzero(&pf_status, sizeof(struct pf_status));
			pf_status.running = 1;
			pf_status.states = states;
#if defined(__FreeBSD__)
			pf_status.since = time_second;
#else
			pf_status.since = time.tv_sec;
#endif
			if (status_ifp != NULL)
#if defined(__FreeBSD__)
				snprintf(pf_status.ifname, IFNAMSIZ, "%s%d",
				    status_ifp->if_name, status_ifp->if_unit);
#else
				strlcpy(pf_status.ifname, 
				    status_ifp->if_xname, IFNAMSIZ);
#endif
			DPFPRINTF(PF_DEBUG_MISC, ("pf: started\n"));
		}
		break;

	case DIOCSTOP:
		if (!pf_status.running)
			error = ENOENT;
		else {
			pf_status.running = 0;
			DPFPRINTF(PF_DEBUG_MISC, ("pf: stopped\n"));
		}
		break;

	case DIOCBEGINRULES: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_rule *rule;

		while ((rule = TAILQ_FIRST(pf_rules_inactive)) != NULL) {
			TAILQ_REMOVE(pf_rules_inactive, rule, entries);
			pf_dynaddr_remove(&rule->src.addr);
			pf_dynaddr_remove(&rule->dst.addr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rule_pl, rule);
#else
			pool_put(&pf_rule_pl, rule);
#endif
		}
		*ticket = ++ticket_rules_inactive;
		break;
	}

	case DIOCADDRULE: {
		struct pfioc_rule *pr = (struct pfioc_rule *)addr;
		struct pf_rule *rule, *tail;

		if (pr->ticket != ticket_rules_inactive) {
			error = EBUSY;
			break;
		}
#if defined(__FreeBSD__)
		rule = uma_zalloc(pf_rule_pl, M_NOWAIT);
#else
		rule = pool_get(&pf_rule_pl, PR_NOWAIT);
#endif
		if (rule == NULL) {
			error = ENOMEM;
			break;
		}
		bcopy(&pr->rule, rule, sizeof(struct pf_rule));
#ifndef INET
		if (rule->af == AF_INET) {
#if defined(__FreeBSD__)
			uma_zfree(pf_rule_pl, rule);
#else
			pool_put(&pf_rule_pl, rule);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET */
#ifndef INET6
		if (rule->af == AF_INET6) {
#if defined(__FreeBSD__)
			uma_zfree(pf_rule_pl, rule);
#else
			pool_put(&pf_rule_pl, rule);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET6 */
		tail = TAILQ_LAST(pf_rules_inactive, pf_rulequeue);
		if (tail)
			rule->nr = tail->nr + 1;
		else
			rule->nr = 0;
		if (rule->ifname[0]) {
			rule->ifp = ifunit(rule->ifname);
			if (rule->ifp == NULL) {
#if defined(__FreeBSD__)
				uma_zfree(pf_rule_pl, rule);
#else
				pool_put(&pf_rule_pl, rule);
#endif
				error = EINVAL;
				break;
			}
		} else
			rule->ifp = NULL;
		if (rule->rt_ifname[0]) {
			rule->rt_ifp = ifunit(rule->rt_ifname);
			if (rule->rt_ifname == NULL) {
#if defined(__FreeBSD__)
				uma_zfree(pf_rule_pl, rule);
#else
				pool_put(&pf_rule_pl, rule);
#endif
				error = EINVAL;
				break;
			}
		} else
			rule->rt_ifp = NULL;
		if (pf_dynaddr_setup(&rule->src.addr, rule->af) ||
		    pf_dynaddr_setup(&rule->dst.addr, rule->af)) {
			pf_dynaddr_remove(&rule->src.addr);
			pf_dynaddr_remove(&rule->dst.addr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rule_pl, rule);
#else
			pool_put(&pf_rule_pl, rule);
#endif
			error = EINVAL;
			break;
		}
		rule->evaluations = rule->packets = rule->bytes = 0;
		TAILQ_INSERT_TAIL(pf_rules_inactive, rule, entries);
		break;
	}

	case DIOCCOMMITRULES: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_rulequeue *old_rules;
		struct pf_rule *rule;
		struct pf_tree_node *n;

		if (*ticket != ticket_rules_inactive) {
			error = EBUSY;
			break;
		}

		/* Swap rules, keep the old. */
		s = splsoftnet();
		/*
		 * Rules are about to get freed, clear rule pointers in states
		 */
		RB_FOREACH(n, pf_state_tree, &tree_ext_gwy)
			n->state->rule.ptr = NULL;
		old_rules = pf_rules_active;
		pf_rules_active = pf_rules_inactive;
		pf_rules_inactive = old_rules;
		ticket_rules_active = ticket_rules_inactive;
		pf_calc_skip_steps(pf_rules_active);
		splx(s);

		/* Purge the old rule list. */
		while ((rule = TAILQ_FIRST(old_rules)) != NULL) {
			TAILQ_REMOVE(old_rules, rule, entries);
			pf_dynaddr_remove(&rule->src.addr);
			pf_dynaddr_remove(&rule->dst.addr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rule_pl, rule);
#else
			pool_put(&pf_rule_pl, rule);
#endif
		}
		break;
	}

	case DIOCGETRULES: {
		struct pfioc_rule *pr = (struct pfioc_rule *)addr;
		struct pf_rule *tail;

		s = splsoftnet();
		tail = TAILQ_LAST(pf_rules_active, pf_rulequeue);
		if (tail)
			pr->nr = tail->nr + 1;
		else
			pr->nr = 0;
		pr->ticket = ticket_rules_active;
		splx(s);
		break;
	}

	case DIOCGETRULE: {
		struct pfioc_rule *pr = (struct pfioc_rule *)addr;
		struct pf_rule *rule;

		if (pr->ticket != ticket_rules_active) {
			error = EBUSY;
			break;
		}
		s = splsoftnet();
		rule = TAILQ_FIRST(pf_rules_active);
		while ((rule != NULL) && (rule->nr != pr->nr))
			rule = TAILQ_NEXT(rule, entries);
		if (rule == NULL) {
			error = EBUSY;
			splx(s);
			break;
		}
		bcopy(rule, &pr->rule, sizeof(struct pf_rule));
		pf_dynaddr_copyout(&pr->rule.src.addr);
		pf_dynaddr_copyout(&pr->rule.dst.addr);
		splx(s);
		break;
	}

	case DIOCCHANGERULE: {
		struct pfioc_changerule *pcr = (struct pfioc_changerule *)addr;
		struct pf_rule *oldrule = NULL, *newrule = NULL;
		u_int32_t nr = 0;

		if (pcr->action < PF_CHANGE_ADD_HEAD ||
		    pcr->action > PF_CHANGE_REMOVE) {
			error = EINVAL;
			break;
		}

		if (pcr->action != PF_CHANGE_REMOVE) {
#if defined(__FreeBSD__)
			newrule = uma_zalloc(pf_rule_pl, M_NOWAIT);
#else
			newrule = pool_get(&pf_rule_pl, PR_NOWAIT);
#endif
			if (newrule == NULL) {
				error = ENOMEM;
				break;
			}
			bcopy(&pcr->newrule, newrule, sizeof(struct pf_rule));
#ifndef INET
			if (newrule->af == AF_INET) {
#if defined(__FreeBSD__)
				uma_zfree(pf_rule_pl, newrule);
#else
				pool_put(&pf_rule_pl, newrule);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET */
#ifndef INET6
			if (newrule->af == AF_INET6) {
#if defined(__FreeBSD__)
				uma_zfree(pf_rule_pl, newrule);
#else
				pool_put(&pf_rule_pl, newrule);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET6 */
			if (newrule->ifname[0]) {
				newrule->ifp = ifunit(newrule->ifname);
				if (newrule->ifp == NULL) {
#if defined(__FreeBSD__)
					uma_zfree(pf_rule_pl, newrule);
#else
					pool_put(&pf_rule_pl, newrule);
#endif
					error = EINVAL;
					break;
				}
			} else
				newrule->ifp = NULL;
			if (newrule->rt_ifname[0]) {
				newrule->rt_ifp = ifunit(newrule->rt_ifname);
				if (newrule->rt_ifname == NULL) {
#if defined(__FreeBSD__)
					uma_zfree(pf_rule_pl, newrule);
#else
					pool_put(&pf_rule_pl, newrule);
#endif
					error = EINVAL;
					break;
				}
			} else
				newrule->rt_ifp = NULL;
			if (pf_dynaddr_setup(&newrule->src.addr, newrule->af) ||
			    pf_dynaddr_setup(&newrule->dst.addr, newrule->af)) {
				pf_dynaddr_remove(&newrule->src.addr);
				pf_dynaddr_remove(&newrule->dst.addr);
#if defined(__FreeBSD__)
				uma_zfree(pf_rule_pl, newrule);
#else
				pool_put(&pf_rule_pl, newrule);
#endif
				error = EINVAL;
				break;
			}
			newrule->evaluations = newrule->packets = 0;
			newrule->bytes = 0;
		}

		s = splsoftnet();

		if (pcr->action == PF_CHANGE_ADD_HEAD)
			oldrule = TAILQ_FIRST(pf_rules_active);
		else if (pcr->action == PF_CHANGE_ADD_TAIL)
			oldrule = TAILQ_LAST(pf_rules_active, pf_rulequeue);
		else {
			oldrule = TAILQ_FIRST(pf_rules_active);
			while ((oldrule != NULL) && pf_compare_rules(oldrule,
			    &pcr->oldrule))
				oldrule = TAILQ_NEXT(oldrule, entries);
			if (oldrule == NULL) {
				error = EINVAL;
				splx(s);
				break;
			}
		}

		if (pcr->action == PF_CHANGE_REMOVE) {
			struct pf_tree_node *n;

			RB_FOREACH(n, pf_state_tree, &tree_ext_gwy)
				if (n->state->rule.ptr == oldrule)
					n->state->rule.ptr = NULL;
			TAILQ_REMOVE(pf_rules_active, oldrule, entries);
			pf_dynaddr_remove(&oldrule->src.addr);
			pf_dynaddr_remove(&oldrule->dst.addr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rule_pl, oldrule);
#else
			pool_put(&pf_rule_pl, oldrule);
#endif
		} else {
			if (oldrule == NULL)
				TAILQ_INSERT_TAIL(pf_rules_active, newrule,
				    entries);
			else if (pcr->action == PF_CHANGE_ADD_HEAD ||
			    pcr->action == PF_CHANGE_ADD_BEFORE)
				TAILQ_INSERT_BEFORE(oldrule, newrule, entries);
			else
				TAILQ_INSERT_AFTER(pf_rules_active, oldrule,
				    newrule, entries);
		}

		TAILQ_FOREACH(oldrule, pf_rules_active, entries)
			oldrule->nr = nr++;

		pf_calc_skip_steps(pf_rules_active);

		ticket_rules_active++;
		splx(s);
		break;
	}

	case DIOCBEGINNATS: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_nat *nat;

		while ((nat = TAILQ_FIRST(pf_nats_inactive)) != NULL) {
			pf_dynaddr_remove(&nat->src.addr);
			pf_dynaddr_remove(&nat->dst.addr);
			pf_dynaddr_remove(&nat->raddr);
			TAILQ_REMOVE(pf_nats_inactive, nat, entries);
#if defined(__FreeBSD__)
			uma_zfree(pf_nat_pl, nat);
#else
			pool_put(&pf_nat_pl, nat);
#endif
		}
		*ticket = ++ticket_nats_inactive;
		break;
	}

	case DIOCADDNAT: {
		struct pfioc_nat *pn = (struct pfioc_nat *)addr;
		struct pf_nat *nat;

		if (pn->ticket != ticket_nats_inactive) {
			error = EBUSY;
			break;
		}
#if defined(__FreeBSD__)
		nat = uma_zalloc(pf_nat_pl, M_NOWAIT);
#else
		nat = pool_get(&pf_nat_pl, PR_NOWAIT);
#endif
		if (nat == NULL) {
			error = ENOMEM;
			break;
		}
		bcopy(&pn->nat, nat, sizeof(struct pf_nat));
#ifndef INET
		if (nat->af == AF_INET) {
#if defined(__FreeBSD__)
			uma_zfree(pf_nat_pl, nat);
#else
			pool_put(&pf_nat_pl, nat);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET */
#ifndef INET6
		if (nat->af == AF_INET6) {
#if defined(__FreeBSD__)
			uma_zfree(pf_nat_pl, nat);
#else
			pool_put(&pf_nat_pl, nat);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET6 */
		if (nat->ifname[0]) {
			nat->ifp = ifunit(nat->ifname);
			if (nat->ifp == NULL) {
#if defined(__FreeBSD__)
				uma_zfree(pf_nat_pl, nat);
#else
				pool_put(&pf_nat_pl, nat);
#endif
				error = EINVAL;
				break;
			}
		} else
			nat->ifp = NULL;
		if (pf_dynaddr_setup(&nat->src.addr, nat->af) ||
		    pf_dynaddr_setup(&nat->dst.addr, nat->af) ||
		    pf_dynaddr_setup(&nat->raddr, nat->af)) {
			pf_dynaddr_remove(&nat->src.addr);
			pf_dynaddr_remove(&nat->dst.addr);
			pf_dynaddr_remove(&nat->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_nat_pl, nat);
#else
			pool_put(&pf_nat_pl, nat);
#endif
			error = EINVAL;
			break;
		}
		TAILQ_INSERT_TAIL(pf_nats_inactive, nat, entries);
		break;
	}

	case DIOCCOMMITNATS: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_natqueue *old_nats;
		struct pf_nat *nat;

		if (*ticket != ticket_nats_inactive) {
			error = EBUSY;
			break;
		}

		/* Swap nats, keep the old. */
		s = splsoftnet();
		old_nats = pf_nats_active;
		pf_nats_active = pf_nats_inactive;
		pf_nats_inactive = old_nats;
		ticket_nats_active = ticket_nats_inactive;
		splx(s);

		/* Purge the old nat list */
		while ((nat = TAILQ_FIRST(old_nats)) != NULL) {
			pf_dynaddr_remove(&nat->src.addr);
			pf_dynaddr_remove(&nat->dst.addr);
			pf_dynaddr_remove(&nat->raddr);
			TAILQ_REMOVE(old_nats, nat, entries);
#if defined(__FreeBSD__)
			uma_zfree(pf_nat_pl, nat);
#else
			pool_put(&pf_nat_pl, nat);
#endif
		}
		break;
	}

	case DIOCGETNATS: {
		struct pfioc_nat *pn = (struct pfioc_nat *)addr;
		struct pf_nat *nat;

		pn->nr = 0;
		s = splsoftnet();
		TAILQ_FOREACH(nat, pf_nats_active, entries)
			pn->nr++;
		pn->ticket = ticket_nats_active;
		splx(s);
		break;
	}

	case DIOCGETNAT: {
		struct pfioc_nat *pn = (struct pfioc_nat *)addr;
		struct pf_nat *nat;
		u_int32_t nr;

		if (pn->ticket != ticket_nats_active) {
			error = EBUSY;
			break;
		}
		nr = 0;
		s = splsoftnet();
		nat = TAILQ_FIRST(pf_nats_active);
		while ((nat != NULL) && (nr < pn->nr)) {
			nat = TAILQ_NEXT(nat, entries);
			nr++;
		}
		if (nat == NULL) {
			error = EBUSY;
			splx(s);
			break;
		}
		bcopy(nat, &pn->nat, sizeof(struct pf_nat));
		pf_dynaddr_copyout(&pn->nat.src.addr);
		pf_dynaddr_copyout(&pn->nat.dst.addr);
		pf_dynaddr_copyout(&pn->nat.raddr);
		splx(s);
		break;
	}

	case DIOCCHANGENAT: {
		struct pfioc_changenat *pcn = (struct pfioc_changenat *)addr;
		struct pf_nat *oldnat = NULL, *newnat = NULL;

		if (pcn->action < PF_CHANGE_ADD_HEAD ||
		    pcn->action > PF_CHANGE_REMOVE) {
			error = EINVAL;
			break;
		}

		if (pcn->action != PF_CHANGE_REMOVE) {
#if defined(__FreeBSD__)
			newnat = uma_zalloc(pf_nat_pl, M_NOWAIT);
#else
			newnat = pool_get(&pf_nat_pl, PR_NOWAIT);
#endif
			if (newnat == NULL) {
				error = ENOMEM;
				break;
			}
			bcopy(&pcn->newnat, newnat, sizeof(struct pf_nat));
#ifndef INET
			if (newnat->af == AF_INET) {
#if defined(__FreeBSD__)
				uma_zfree(pf_nat_pl, newnat);
#else
				pool_put(&pf_nat_pl, newnat);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET */
#ifndef INET6
			if (newnat->af == AF_INET6) {
#if defined(__FreeBSD__)
				uma_zfree(pf_nat_pl, newnat);
#else
				pool_put(&pf_nat_pl, newnat);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET6 */
			if (newnat->ifname[0]) {
				newnat->ifp = ifunit(newnat->ifname);
				if (newnat->ifp == NULL) {
#if defined(__FreeBSD__)
					uma_zfree(pf_nat_pl, newnat);
#else
					pool_put(&pf_nat_pl, newnat);
#endif
					error = EINVAL;
					break;
				}
			} else
				newnat->ifp = NULL;
			if (pf_dynaddr_setup(&newnat->src.addr, newnat->af) ||
			    pf_dynaddr_setup(&newnat->dst.addr, newnat->af) ||
			    pf_dynaddr_setup(&newnat->raddr, newnat->af)) {
				pf_dynaddr_remove(&newnat->src.addr);
				pf_dynaddr_remove(&newnat->dst.addr);
				pf_dynaddr_remove(&newnat->raddr);
#if defined(__FreeBSD__)
				uma_zfree(pf_nat_pl, newnat);
#else
				pool_put(&pf_nat_pl, newnat);
#endif
				error = EINVAL;
				break;
			}
		}

		s = splsoftnet();

		if (pcn->action == PF_CHANGE_ADD_HEAD)
			oldnat = TAILQ_FIRST(pf_nats_active);
		else if (pcn->action == PF_CHANGE_ADD_TAIL)
			oldnat = TAILQ_LAST(pf_nats_active, pf_natqueue);
		else {
			oldnat = TAILQ_FIRST(pf_nats_active);
			while ((oldnat != NULL) && pf_compare_nats(oldnat,
			    &pcn->oldnat))
				oldnat = TAILQ_NEXT(oldnat, entries);
			if (oldnat == NULL) {
				error = EINVAL;
				splx(s);
				break;
			}
		}

		if (pcn->action == PF_CHANGE_REMOVE) {
			pf_dynaddr_remove(&oldnat->src.addr);
			pf_dynaddr_remove(&oldnat->dst.addr);
			pf_dynaddr_remove(&oldnat->raddr);
			TAILQ_REMOVE(pf_nats_active, oldnat, entries);
#if defined(__FreeBSD__)
			uma_zfree(pf_nat_pl, oldnat);
#else
			pool_put(&pf_nat_pl, oldnat);
#endif
		} else {
			if (oldnat == NULL)
				TAILQ_INSERT_TAIL(pf_nats_active, newnat,
				    entries);
			else if (pcn->action == PF_CHANGE_ADD_HEAD ||
			    pcn->action == PF_CHANGE_ADD_BEFORE)
				TAILQ_INSERT_BEFORE(oldnat, newnat, entries);
			else
				TAILQ_INSERT_AFTER(pf_nats_active, oldnat,
				    newnat, entries);
		}

		ticket_nats_active++;
		splx(s);
		break;
	}

	case DIOCBEGINBINATS: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_binat *binat;

		while ((binat = TAILQ_FIRST(pf_binats_inactive)) != NULL) {
			TAILQ_REMOVE(pf_binats_inactive, binat, entries);
			pf_dynaddr_remove(&binat->saddr);
			pf_dynaddr_remove(&binat->daddr);
			pf_dynaddr_remove(&binat->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_binat_pl, binat);
#else
			pool_put(&pf_binat_pl, binat);
#endif
		}
		*ticket = ++ticket_binats_inactive;
		break;
	}

	case DIOCADDBINAT: {
		struct pfioc_binat *pb = (struct pfioc_binat *)addr;
		struct pf_binat *binat;

		if (pb->ticket != ticket_binats_inactive) {
			error = EBUSY;
			break;
		}
#if defined(__FreeBSD__)
		binat = uma_zalloc(pf_binat_pl, M_NOWAIT);
#else
		binat = pool_get(&pf_binat_pl, PR_NOWAIT);
#endif
		if (binat == NULL) {
			error = ENOMEM;
			break;
		}
		bcopy(&pb->binat, binat, sizeof(struct pf_binat));
#ifndef INET
		if (binat->af == AF_INET) {
#if defined(__FreeBSD__)
			uma_zfree(pf_binat_pl, binat);
#else
			pool_put(&pf_binat_pl, binat);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET */
#ifndef INET6
		if (binat->af == AF_INET6) {
#if defined(__FreeBSD__)
			uma_zfree(pf_binat_pl, binat);
#else
			pool_put(&pf_binat_pl, binat);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET6 */
		if (binat->ifname[0]) {
			binat->ifp = ifunit(binat->ifname);
			if (binat->ifp == NULL) {
#if defined(__FreeBSD__)
				uma_zfree(pf_binat_pl, binat);
#else
				pool_put(&pf_binat_pl, binat);
#endif
				error = EINVAL;
				break;
			}
		} else
			binat->ifp = NULL;
		if (pf_dynaddr_setup(&binat->saddr, binat->af) ||
		    pf_dynaddr_setup(&binat->daddr, binat->af) ||
		    pf_dynaddr_setup(&binat->raddr, binat->af)) {
			pf_dynaddr_remove(&binat->saddr);
			pf_dynaddr_remove(&binat->daddr);
			pf_dynaddr_remove(&binat->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_binat_pl, binat);
#else
			pool_put(&pf_binat_pl, binat);
#endif
			error = EINVAL;
			break;
		}
		TAILQ_INSERT_TAIL(pf_binats_inactive, binat, entries);
		break;
	}

	case DIOCCOMMITBINATS: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_binatqueue *old_binats;
		struct pf_binat *binat;

		if (*ticket != ticket_binats_inactive) {
			error = EBUSY;
			break;
		}

		/* Swap binats, keep the old. */
		s = splsoftnet();
		old_binats = pf_binats_active;
		pf_binats_active = pf_binats_inactive;
		pf_binats_inactive = old_binats;
		ticket_binats_active = ticket_binats_inactive;
		splx(s);

		/* Purge the old binat list */
		while ((binat = TAILQ_FIRST(old_binats)) != NULL) {
			TAILQ_REMOVE(old_binats, binat, entries);
			pf_dynaddr_remove(&binat->saddr);
			pf_dynaddr_remove(&binat->daddr);
			pf_dynaddr_remove(&binat->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_binat_pl, binat);
#else
			pool_put(&pf_binat_pl, binat);
#endif
		}
		break;
	}

	case DIOCGETBINATS: {
		struct pfioc_binat *pb = (struct pfioc_binat *)addr;
		struct pf_binat *binat;

		pb->nr = 0;
		s = splsoftnet();
		TAILQ_FOREACH(binat, pf_binats_active, entries)
			pb->nr++;
		pb->ticket = ticket_binats_active;
		splx(s);
		break;
	}

	case DIOCGETBINAT: {
		struct pfioc_binat *pb = (struct pfioc_binat *)addr;
		struct pf_binat *binat;
		u_int32_t nr;

		if (pb->ticket != ticket_binats_active) {
			error = EBUSY;
			break;
		}
		nr = 0;
		s = splsoftnet();
		binat = TAILQ_FIRST(pf_binats_active);
		while ((binat != NULL) && (nr < pb->nr)) {
			binat = TAILQ_NEXT(binat, entries);
			nr++;
		}
		if (binat == NULL) {
			error = EBUSY;
			splx(s);
			break;
		}
		bcopy(binat, &pb->binat, sizeof(struct pf_binat));
		pf_dynaddr_copyout(&pb->binat.saddr);
		pf_dynaddr_copyout(&pb->binat.daddr);
		pf_dynaddr_copyout(&pb->binat.raddr);
		splx(s);
		break;
	}

	case DIOCCHANGEBINAT: {
		struct pfioc_changebinat *pcn = (struct pfioc_changebinat *)addr;
		struct pf_binat *oldbinat = NULL, *newbinat = NULL;

		if (pcn->action < PF_CHANGE_ADD_HEAD ||
		    pcn->action > PF_CHANGE_REMOVE) {
			error = EINVAL;
			break;
		}

		if (pcn->action != PF_CHANGE_REMOVE) {
#if defined(__FreeBSD__)
			newbinat = uma_zalloc(pf_binat_pl, M_NOWAIT);
#else
			newbinat = pool_get(&pf_binat_pl, PR_NOWAIT);
#endif
			if (newbinat == NULL) {
				error = ENOMEM;
				break;
			}
			bcopy(&pcn->newbinat, newbinat,
				sizeof(struct pf_binat));
#ifndef INET
			if (newbinat->af == AF_INET) {
#if defined(__FreeBSD__)
				uma_zfree(pf_binat_pl, newbinat);
#else
				pool_put(&pf_binat_pl, newbinat);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET */
#ifndef INET6
			if (newbinat->af == AF_INET6) {
#if defined(__FreeBSD__)
				uma_zfree(pf_binat_pl, newbinat);
#else
				pool_put(&pf_binat_pl, newbinat);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET6 */
			if (newbinat->ifname[0]) {
				newbinat->ifp = ifunit(newbinat->ifname);
				if (newbinat->ifp == NULL) {
#if defined(__FreeBSD__)
					uma_zfree(pf_binat_pl, newbinat);
#else
					pool_put(&pf_binat_pl, newbinat);
#endif
					error = EINVAL;
					break;
				}
			} else
				newbinat->ifp = NULL;
			if (pf_dynaddr_setup(&newbinat->saddr, newbinat->af) ||
			    pf_dynaddr_setup(&newbinat->daddr, newbinat->af) ||
			    pf_dynaddr_setup(&newbinat->raddr, newbinat->af)) {
				pf_dynaddr_remove(&newbinat->saddr);
				pf_dynaddr_remove(&newbinat->daddr);
				pf_dynaddr_remove(&newbinat->raddr);
#if defined(__FreeBSD__)
				uma_zfree(pf_binat_pl, newbinat);
#else
				pool_put(&pf_binat_pl, newbinat);
#endif
				error = EINVAL;
				break;
			}
		}

		s = splsoftnet();

		if (pcn->action == PF_CHANGE_ADD_HEAD)
			oldbinat = TAILQ_FIRST(pf_binats_active);
		else if (pcn->action == PF_CHANGE_ADD_TAIL)
			oldbinat = TAILQ_LAST(pf_binats_active, pf_binatqueue);
		else {
			oldbinat = TAILQ_FIRST(pf_binats_active);
			while ((oldbinat != NULL) && pf_compare_binats(oldbinat,
			    &pcn->oldbinat))
				oldbinat = TAILQ_NEXT(oldbinat, entries);
			if (oldbinat == NULL) {
				error = EINVAL;
				splx(s);
				break;
			}
		}

		if (pcn->action == PF_CHANGE_REMOVE) {
			TAILQ_REMOVE(pf_binats_active, oldbinat, entries);
			pf_dynaddr_remove(&oldbinat->saddr);
			pf_dynaddr_remove(&oldbinat->daddr);
			pf_dynaddr_remove(&oldbinat->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_binat_pl, oldbinat);
#else
			pool_put(&pf_binat_pl, oldbinat);
#endif
		} else {
			if (oldbinat == NULL)
				TAILQ_INSERT_TAIL(pf_binats_active, newbinat,
				    entries);
			else if (pcn->action == PF_CHANGE_ADD_HEAD ||
			    pcn->action == PF_CHANGE_ADD_BEFORE)
				TAILQ_INSERT_BEFORE(oldbinat, newbinat,
				    entries);
			else
				TAILQ_INSERT_AFTER(pf_binats_active, oldbinat,
				    newbinat, entries);
		}

		ticket_binats_active++;
		splx(s);
		break;
	}

	case DIOCBEGINRDRS: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_rdr *rdr;

		while ((rdr = TAILQ_FIRST(pf_rdrs_inactive)) != NULL) {
			TAILQ_REMOVE(pf_rdrs_inactive, rdr, entries);
			pf_dynaddr_remove(&rdr->saddr);
			pf_dynaddr_remove(&rdr->daddr);
			pf_dynaddr_remove(&rdr->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rdr_pl, rdr);
#else
			pool_put(&pf_rdr_pl, rdr);
#endif
		}
		*ticket = ++ticket_rdrs_inactive;
		break;
	}

	case DIOCADDRDR: {
		struct pfioc_rdr *pr = (struct pfioc_rdr *)addr;
		struct pf_rdr *rdr;

		if (pr->ticket != ticket_rdrs_inactive) {
			error = EBUSY;
			break;
		}
#if defined(__FreeBSD__)
		rdr = uma_zalloc(pf_rdr_pl, M_NOWAIT);
#else
		rdr = pool_get(&pf_rdr_pl, PR_NOWAIT);
#endif
		if (rdr == NULL) {
			error = ENOMEM;
			break;
		}
		bcopy(&pr->rdr, rdr, sizeof(struct pf_rdr));
#ifndef INET
		if (rdr->af == AF_INET) {
#if defined(__FreeBSD__)
			uma_zfree(pf_rdr_pl, rdr);
#else
			pool_put(&pf_rdr_pl, rdr);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET */
#ifndef INET6
		if (rdr->af == AF_INET6) {
#if defined(__FreeBSD__)
			uma_zfree(pf_rdr_pl, rdr);
#else
			pool_put(&pf_rdr_pl, rdr);
#endif
			error = EAFNOSUPPORT;
			break;
		}
#endif /* INET6 */
		if (rdr->ifname[0]) {
			rdr->ifp = ifunit(rdr->ifname);
			if (rdr->ifp == NULL) {
#if defined(__FreeBSD__)
				uma_zfree(pf_rdr_pl, rdr);
#else
				pool_put(&pf_rdr_pl, rdr);
#endif
				error = EINVAL;
				break;
			}
		} else
			rdr->ifp = NULL;
		if (pf_dynaddr_setup(&rdr->saddr, rdr->af) ||
		    pf_dynaddr_setup(&rdr->daddr, rdr->af) ||
		    pf_dynaddr_setup(&rdr->raddr, rdr->af)) {
			pf_dynaddr_remove(&rdr->saddr);
			pf_dynaddr_remove(&rdr->daddr);
			pf_dynaddr_remove(&rdr->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rdr_pl, rdr);
#else
			pool_put(&pf_rdr_pl, rdr);
#endif
			error = EINVAL;
			break;
		}
		TAILQ_INSERT_TAIL(pf_rdrs_inactive, rdr, entries);
		break;
	}

	case DIOCCOMMITRDRS: {
		u_int32_t *ticket = (u_int32_t *)addr;
		struct pf_rdrqueue *old_rdrs;
		struct pf_rdr *rdr;

		if (*ticket != ticket_rdrs_inactive) {
			error = EBUSY;
			break;
		}

		/* Swap rdrs, keep the old. */
		s = splsoftnet();
		old_rdrs = pf_rdrs_active;
		pf_rdrs_active = pf_rdrs_inactive;
		pf_rdrs_inactive = old_rdrs;
		ticket_rdrs_active = ticket_rdrs_inactive;
		splx(s);

		/* Purge the old rdr list */
		while ((rdr = TAILQ_FIRST(old_rdrs)) != NULL) {
			TAILQ_REMOVE(old_rdrs, rdr, entries);
			pf_dynaddr_remove(&rdr->saddr);
			pf_dynaddr_remove(&rdr->daddr);
			pf_dynaddr_remove(&rdr->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rdr_pl, rdr);
#else
			pool_put(&pf_rdr_pl, rdr);
#endif
		}
		break;
	}

	case DIOCGETRDRS: {
		struct pfioc_rdr *pr = (struct pfioc_rdr *)addr;
		struct pf_rdr *rdr;

		pr->nr = 0;
		s = splsoftnet();
		TAILQ_FOREACH(rdr, pf_rdrs_active, entries)
			pr->nr++;
		pr->ticket = ticket_rdrs_active;
		splx(s);
		break;
	}

	case DIOCGETRDR: {
		struct pfioc_rdr *pr = (struct pfioc_rdr *)addr;
		struct pf_rdr *rdr;
		u_int32_t nr;

		if (pr->ticket != ticket_rdrs_active) {
			error = EBUSY;
			break;
		}
		nr = 0;
		s = splsoftnet();
		rdr = TAILQ_FIRST(pf_rdrs_active);
		while ((rdr != NULL) && (nr < pr->nr)) {
			rdr = TAILQ_NEXT(rdr, entries);
			nr++;
		}
		if (rdr == NULL) {
			error = EBUSY;
			splx(s);
			break;
		}
		bcopy(rdr, &pr->rdr, sizeof(struct pf_rdr));
		pf_dynaddr_copyout(&pr->rdr.saddr);
		pf_dynaddr_copyout(&pr->rdr.daddr);
		pf_dynaddr_copyout(&pr->rdr.raddr);
		splx(s);
		break;
	}

	case DIOCCHANGERDR: {
		struct pfioc_changerdr *pcn = (struct pfioc_changerdr *)addr;
		struct pf_rdr *oldrdr = NULL, *newrdr = NULL;

		if (pcn->action < PF_CHANGE_ADD_HEAD ||
		    pcn->action > PF_CHANGE_REMOVE) {
			error = EINVAL;
			break;
		}

		if (pcn->action != PF_CHANGE_REMOVE) {
#if defined(__FreeBSD__)
			newrdr = uma_zalloc(pf_rdr_pl, M_NOWAIT);
#else
			newrdr = pool_get(&pf_rdr_pl, PR_NOWAIT);
#endif
			if (newrdr == NULL) {
				error = ENOMEM;
				break;
			}
			bcopy(&pcn->newrdr, newrdr, sizeof(struct pf_rdr));
#ifndef INET
			if (newrdr->af == AF_INET) {
#if defined(__FreeBSD__)
				uma_zfree(pf_rdr_pl, newrdr);
#else
				pool_put(&pf_rdr_pl, newrdr);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET */
#ifndef INET6
			if (newrdr->af == AF_INET6) {
#if defined(__FreeBSD__)
				uma_zfree(pf_rdr_pl, newrdr);
#else
				pool_put(&pf_rdr_pl, newrdr);
#endif
				error = EAFNOSUPPORT;
				break;
			}
#endif /* INET6 */
			if (newrdr->ifname[0]) {
				newrdr->ifp = ifunit(newrdr->ifname);
				if (newrdr->ifp == NULL) {
#if defined(__FreeBSD__)
					uma_zfree(pf_rdr_pl, newrdr);
#else
					pool_put(&pf_rdr_pl, newrdr);
#endif
					error = EINVAL;
					break;
				}
			} else
				newrdr->ifp = NULL;
			if (pf_dynaddr_setup(&newrdr->saddr, newrdr->af) ||
			    pf_dynaddr_setup(&newrdr->daddr, newrdr->af) ||
			    pf_dynaddr_setup(&newrdr->raddr, newrdr->af)) {
				pf_dynaddr_remove(&newrdr->saddr);
				pf_dynaddr_remove(&newrdr->daddr);
				pf_dynaddr_remove(&newrdr->raddr);
#if defined(__FreeBSD__)
				uma_zfree(pf_rdr_pl, newrdr);
#else
				pool_put(&pf_rdr_pl, newrdr);
#endif
				error = EINVAL;
				break;
			}
		}

		s = splsoftnet();

		if (pcn->action == PF_CHANGE_ADD_HEAD)
			oldrdr = TAILQ_FIRST(pf_rdrs_active);
		else if (pcn->action == PF_CHANGE_ADD_TAIL)
			oldrdr = TAILQ_LAST(pf_rdrs_active, pf_rdrqueue);
		else {
			oldrdr = TAILQ_FIRST(pf_rdrs_active);
			while ((oldrdr != NULL) && pf_compare_rdrs(oldrdr,
			    &pcn->oldrdr))
				oldrdr = TAILQ_NEXT(oldrdr, entries);
			if (oldrdr == NULL) {
				error = EINVAL;
				splx(s);
				break;
			}
		}

		if (pcn->action == PF_CHANGE_REMOVE) {
			TAILQ_REMOVE(pf_rdrs_active, oldrdr, entries);
			pf_dynaddr_remove(&oldrdr->saddr);
			pf_dynaddr_remove(&oldrdr->daddr);
			pf_dynaddr_remove(&oldrdr->raddr);
#if defined(__FreeBSD__)
			uma_zfree(pf_rdr_pl, oldrdr);
#else
			pool_put(&pf_rdr_pl, oldrdr);
#endif
		} else {
			if (oldrdr == NULL)
				TAILQ_INSERT_TAIL(pf_rdrs_active, newrdr,
				    entries);
			else if (pcn->action == PF_CHANGE_ADD_HEAD ||
			    pcn->action == PF_CHANGE_ADD_BEFORE)
				TAILQ_INSERT_BEFORE(oldrdr, newrdr, entries);
			else
				TAILQ_INSERT_AFTER(pf_rdrs_active, oldrdr,
				    newrdr, entries);
		}

		ticket_rdrs_active++;
		splx(s);
		break;
	}

	case DIOCCLRSTATES: {
		struct pf_tree_node *n;

		s = splsoftnet();
		RB_FOREACH(n, pf_state_tree, &tree_ext_gwy)
			n->state->expire = 0;
		pf_purge_expired_states();
		pf_status.states = 0;
		splx(s);
		break;
	}

	case DIOCKILLSTATES: {
		struct pf_tree_node *n;
		struct pf_state *st;
		struct pfioc_state_kill *psk =
		    (struct pfioc_state_kill *)addr;
		int killed = 0;

		s = splsoftnet();
		RB_FOREACH(n, pf_state_tree, &tree_ext_gwy) {
			st = n->state;
			if ((!psk->psk_af || st->af == psk->psk_af) &&
			    (!psk->psk_proto || psk->psk_proto == st->proto) &&
			    PF_MATCHA(psk->psk_src.not, &psk->psk_src.addr.addr,
			    &psk->psk_src.mask, &st->lan.addr, st->af) &&
			    PF_MATCHA(psk->psk_dst.not, &psk->psk_dst.addr.addr,
			    &psk->psk_dst.mask, &st->ext.addr, st->af) &&
			    (psk->psk_src.port_op == 0 ||
			    pf_match_port(psk->psk_src.port_op,
			    psk->psk_src.port[0], psk->psk_src.port[1],
			    st->lan.port)) &&
			    (psk->psk_dst.port_op == 0 ||
			    pf_match_port(psk->psk_dst.port_op,
			    psk->psk_dst.port[0], psk->psk_dst.port[1],
			    st->ext.port))) {
				st->expire = 0;
				killed++;
			}
		}
		pf_purge_expired_states();
		splx(s);
		psk->psk_af = killed;
		break;
	}

	case DIOCADDSTATE: {
		struct pfioc_state *ps = (struct pfioc_state *)addr;
		struct pf_state *state;

#if defined(__FreeBSD__)
		state = uma_zalloc(pf_state_pl, M_NOWAIT);
#else
		state = pool_get(&pf_state_pl, PR_NOWAIT);
#endif
		if (state == NULL) {
			error = ENOMEM;
			break;
		}
		s = splsoftnet();
		bcopy(&ps->state, state, sizeof(struct pf_state));
		state->rule.ptr = NULL;
#if defined(__FreeBSD__)
		state->creation = time_second;
#else
		state->creation = time.tv_sec;
#endif
		state->expire += state->creation;
		state->packets = 0;
		state->bytes = 0;
		if (pf_insert_state(state)) {
#if defined(__FreeBSD__)
			uma_zfree(pf_state_pl, state);
#else
			pool_put(&pf_state_pl, state);
#endif
			error = ENOMEM;
		}
		splx(s);
	}

	case DIOCGETSTATE: {
		struct pfioc_state *ps = (struct pfioc_state *)addr;
		struct pf_tree_node *n;
		u_int32_t nr;
		int secs;

		nr = 0;
		s = splsoftnet();
		RB_FOREACH(n, pf_state_tree, &tree_ext_gwy) {
			if (nr >= ps->nr)
				break;
			nr++;
		}
		if (n == NULL) {
			error = EBUSY;
			splx(s);
			break;
		}
		bcopy(n->state, &ps->state, sizeof(struct pf_state));
		if (n->state->rule.ptr == NULL)
			ps->state.rule.nr = -1;
		else
			ps->state.rule.nr = n->state->rule.ptr->nr;
		splx(s);
#if defined(__FreeBSD__)
		secs = time_second;
#else
		secs = time.tv_sec;
#endif
		ps->state.creation = secs - ps->state.creation;
		if (ps->state.expire <= secs)
			ps->state.expire = 0;
		else
			ps->state.expire -= secs;
		break;
	}

	case DIOCGETSTATES: {
		struct pfioc_states *ps = (struct pfioc_states *)addr;
		struct pf_tree_node *n;
		struct pf_state *p, pstore;
		u_int32_t nr = 0;
		int space = ps->ps_len;

		if (space == 0) {
			s = splsoftnet();
			RB_FOREACH(n, pf_state_tree, &tree_ext_gwy)
				nr++;
			splx(s);
			ps->ps_len = sizeof(struct pf_state) * nr;
#if defined(__FreeBSD__)
			mtx_unlock(&pf_rule_mtx);
#endif
			return (0);
		}

		s = splsoftnet();
		p = ps->ps_states;
		RB_FOREACH(n, pf_state_tree, &tree_ext_gwy) {
#if defined(__FreeBSD__)
			int secs = time_second;
#else
			int secs = time.tv_sec;
#endif

			if ((nr + 1) * sizeof(*p) > ps->ps_len)
				break;

			bcopy(n->state, &pstore, sizeof(pstore));
			if (n->state->rule.ptr == NULL)
				pstore.rule.nr = -1;
			else
				pstore.rule.nr = n->state->rule.ptr->nr;
			pstore.creation = secs - pstore.creation;
			if (pstore.expire <= secs)
				pstore.expire = 0;
			else
				pstore.expire -= secs;
			error = copyout(&pstore, p, sizeof(*p));
			if (error) {
				splx(s);
				goto fail;
			}
			p++;
			nr++;
		}
		ps->ps_len = sizeof(struct pf_state) * nr;
		splx(s);
		break;
	}

	case DIOCSETSTATUSIF: {
		struct pfioc_if *pi = (struct pfioc_if *)addr;
		struct ifnet *ifp;

		if (pi->ifname[0] == 0) {
			status_ifp = NULL;
			bzero(pf_status.ifname, IFNAMSIZ);
		} else
			if ((ifp = ifunit(pi->ifname)) == NULL)
				error = EINVAL;
			else {
				status_ifp = ifp;
#if defined(__FreeBSD__)
				snprintf(pf_status.ifname, IFNAMSIZ, "%s%d",
				     ifp->if_name, ifp->if_unit);
#else
				strlcpy(pf_status.ifname,
				    ifp->if_xname, IFNAMSIZ);
#endif
			}
		break;
	}

	case DIOCGETSTATUS: {
		struct pf_status *s = (struct pf_status *)addr;
		bcopy(&pf_status, s, sizeof(struct pf_status));
		break;
	}

	case DIOCCLRSTATUS: {
		u_int32_t running = pf_status.running;
		u_int32_t states = pf_status.states;
		u_int32_t since = pf_status.since;
		u_int32_t debug = pf_status.debug;

		bzero(&pf_status, sizeof(struct pf_status));
		pf_status.running = running;
		pf_status.states = states;
		pf_status.since = since;
		pf_status.debug = debug;
		if (status_ifp != NULL)
#if defined(__FreeBSD__)
			snprintf(pf_status.ifname, IFNAMSIZ, "%s%d",
			    status_ifp->if_name, status_ifp->if_unit);
#else
			strlcpy(pf_status.ifname,
			    status_ifp->if_xname, IFNAMSIZ);
#endif
		break;
	}

	case DIOCNATLOOK: {
		struct pfioc_natlook *pnl = (struct pfioc_natlook *)addr;
		struct pf_state *st;
		struct pf_tree_node key;
		int direction = pnl->direction;

		key.af = pnl->af;
		key.proto = pnl->proto;

		/*
		 * userland gives us source and dest of connetion, reverse
		 * the lookup so we ask for what happens with the return
		 * traffic, enabling us to find it in the state tree.
		 */
		PF_ACPY(&key.addr[1], &pnl->saddr, pnl->af);
		key.port[1] = pnl->sport;
		PF_ACPY(&key.addr[0], &pnl->daddr, pnl->af);
		key.port[0] = pnl->dport;

		if (!pnl->proto ||
		    PF_AZERO(&pnl->saddr, pnl->af) ||
		    PF_AZERO(&pnl->daddr, pnl->af) ||
		    !pnl->dport || !pnl->sport)
			error = EINVAL;
		else {
			s = splsoftnet();
			if (direction == PF_IN)
				st = pf_find_state(&tree_ext_gwy, &key);
			else
				st = pf_find_state(&tree_lan_ext, &key);
			if (st != NULL) {
				if (direction  == PF_IN) {
					PF_ACPY(&pnl->rsaddr, &st->lan.addr,
					    st->af);
					pnl->rsport = st->lan.port;
					PF_ACPY(&pnl->rdaddr, &pnl->daddr,
					    pnl->af);
					pnl->rdport = pnl->dport;
				} else {
					PF_ACPY(&pnl->rdaddr, &st->gwy.addr,
					    st->af);
					pnl->rdport = st->gwy.port;
					PF_ACPY(&pnl->rsaddr, &pnl->saddr,
					    pnl->af);
					pnl->rsport = pnl->sport;
				}
			} else
				error = ENOENT;
			splx(s);
		}
		break;
	}

	case DIOCSETTIMEOUT: {
		struct pfioc_tm *pt = (struct pfioc_tm *)addr;
		int old;

		if (pt->timeout < 0 || pt->timeout >= PFTM_MAX ||
		    pt->seconds < 0) {
			error = EINVAL;
			goto fail;
		}
		old = *pftm_timeouts[pt->timeout];
		*pftm_timeouts[pt->timeout] = pt->seconds;
		pt->seconds = old;
		break;
	}

	case DIOCGETTIMEOUT: {
		struct pfioc_tm *pt = (struct pfioc_tm *)addr;

		if (pt->timeout < 0 || pt->timeout >= PFTM_MAX) {
			error = EINVAL;
			goto fail;
		}
		pt->seconds = *pftm_timeouts[pt->timeout];
		break;
	}

	case DIOCGETLIMIT: {
		struct pfioc_limit *pl = (struct pfioc_limit *)addr;

		if (pl->index < 0 || pl->index >= PF_LIMIT_MAX) {
			error = EINVAL;
			goto fail;
		}
		pl->limit = pf_pool_limits[pl->index].limit;
		break;
	}

	case DIOCSETLIMIT: {
		struct pfioc_limit *pl = (struct pfioc_limit *)addr;
		int old_limit;

		if (pl->index < 0 || pl->index >= PF_LIMIT_MAX) {
			error = EINVAL;
			goto fail;
		}
#if defined(__FreeBSD__)
		uma_zone_set_max(pf_pool_limits[pl->index].pp, pl->limit);
#else
		if (pool_sethardlimit(pf_pool_limits[pl->index],
		    pl->limit, NULL, 0) != 0) {
	 	 	error = EBUSY;
	 	 	goto fail;
		}
#endif
		old_limit = pf_pool_limits[pl->index].limit;
		pf_pool_limits[pl->index].limit = pl->limit;
		pl->limit = old_limit;
		break;
	}

	case DIOCSETDEBUG: {
		u_int32_t *level = (u_int32_t *)addr;
		pf_status.debug = *level;
		break;
	}

	case DIOCCLRRULECTRS: {
		struct pf_rule *rule;

		s = splsoftnet();
		TAILQ_FOREACH(rule, pf_rules_active, entries)
			rule->evaluations = rule->packets =
			    rule->bytes = 0;
		splx(s);
		break;
	}

	default:
		error = ENODEV;
		break;
	}

fail:
	mtx_unlock(&pf_rule_mtx);

	return (error);
}

#if defined(__FreeBSD__)
static int
pf_beginrules(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_rule *rule;

	while ((rule = TAILQ_FIRST(pf_rules_inactive)) != NULL) {
		TAILQ_REMOVE(pf_rules_inactive, rule, entries);
		pf_dynaddr_remove(&rule->src.addr);
		pf_dynaddr_remove(&rule->dst.addr);
		uma_zfree(pf_rule_pl, rule);
	}
	*ticket = ++ticket_rules_inactive;
	return (0);
}

static int
pf_commitrules(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_rulequeue *old_rules;
	struct pf_rule *rule;
	struct pf_tree_node *n;
	int s;

	if (*ticket != ticket_rules_inactive)
		return (1);

	/* Swap rules, keep the old. */
	s = splsoftnet();
	/*
	 * Rules are about to get freed, clear rule pointers in states
	 */
	RB_FOREACH(n, pf_state_tree, &tree_ext_gwy)
		n->state->rule.ptr = NULL;
	old_rules = pf_rules_active;
	pf_rules_active = pf_rules_inactive;
	pf_rules_inactive = old_rules;
	ticket_rules_active = ticket_rules_inactive;
	pf_calc_skip_steps(pf_rules_active);
	splx(s);

	/* Purge the old rule list. */
	while ((rule = TAILQ_FIRST(old_rules)) != NULL) {
		TAILQ_REMOVE(old_rules, rule, entries);
		pf_dynaddr_remove(&rule->src.addr);
		pf_dynaddr_remove(&rule->dst.addr);
		uma_zfree(pf_rule_pl, rule);
	}
	return (0);
}

static int
pf_beginnats(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_nat *nat;

	while ((nat = TAILQ_FIRST(pf_nats_inactive)) != NULL) {
		pf_dynaddr_remove(&nat->src.addr);
		pf_dynaddr_remove(&nat->dst.addr);
		pf_dynaddr_remove(&nat->raddr);
		TAILQ_REMOVE(pf_nats_inactive, nat, entries);
		uma_zfree(pf_nat_pl, nat);
	}
	*ticket = ++ticket_nats_inactive;
	return (0);
}

static int
pf_commitnats(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_natqueue *old_nats;
	struct pf_nat *nat;
	int s;

	if (*ticket != ticket_nats_inactive)
		return (1);

	/* Swap nats, keep the old. */
	s = splsoftnet();
	old_nats = pf_nats_active;
	pf_nats_active = pf_nats_inactive;
	pf_nats_inactive = old_nats;
	ticket_nats_active = ticket_nats_inactive;
	splx(s);

	/* Purge the old nat list */
	while ((nat = TAILQ_FIRST(old_nats)) != NULL) {
		pf_dynaddr_remove(&nat->src.addr);
		pf_dynaddr_remove(&nat->dst.addr);
		pf_dynaddr_remove(&nat->raddr);
		TAILQ_REMOVE(old_nats, nat, entries);
		uma_zfree(pf_nat_pl, nat);
	}
	return (0);
}

static int
pf_beginbinats(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_binat *binat;

	while ((binat = TAILQ_FIRST(pf_binats_inactive)) != NULL) {
		TAILQ_REMOVE(pf_binats_inactive, binat, entries);
		pf_dynaddr_remove(&binat->saddr);
		pf_dynaddr_remove(&binat->daddr);
		pf_dynaddr_remove(&binat->raddr);
		uma_zfree(pf_binat_pl, binat);
	}
	*ticket = ++ticket_binats_inactive;
	return (0);
}

static int
pf_commitbinats(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_binatqueue *old_binats;
	struct pf_binat *binat;
	int s;

	if (*ticket != ticket_binats_inactive)
		return (1);

	/* Swap binats, keep the old. */
	s = splsoftnet();
	old_binats = pf_binats_active;
	pf_binats_active = pf_binats_inactive;
	pf_binats_inactive = old_binats;
	ticket_binats_active = ticket_binats_inactive;
	splx(s);

	/* Purge the old binat list */
	while ((binat = TAILQ_FIRST(old_binats)) != NULL) {
		TAILQ_REMOVE(old_binats, binat, entries);
		pf_dynaddr_remove(&binat->saddr);
		pf_dynaddr_remove(&binat->daddr);
		pf_dynaddr_remove(&binat->raddr);
		uma_zfree(pf_binat_pl, binat);
	}
	return (0);
}

static int
pf_beginrdrs(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_rdr *rdr;

	while ((rdr = TAILQ_FIRST(pf_rdrs_inactive)) != NULL) {
		TAILQ_REMOVE(pf_rdrs_inactive, rdr, entries);
		pf_dynaddr_remove(&rdr->saddr);
		pf_dynaddr_remove(&rdr->daddr);
		pf_dynaddr_remove(&rdr->raddr);
		uma_zfree(pf_rdr_pl, rdr);
	}
	*ticket = ++ticket_rdrs_inactive;
	return (0);
}

static int
pf_commitrdrs(void *addr)
{
	u_int32_t *ticket = (u_int32_t *)addr;
	struct pf_rdrqueue *old_rdrs;
	struct pf_rdr *rdr;
	int s;

	if (*ticket != ticket_rdrs_inactive)
		return (1);

	/* Swap rdrs, keep the old. */
	s = splsoftnet();
	old_rdrs = pf_rdrs_active;
	pf_rdrs_active = pf_rdrs_inactive;
	pf_rdrs_inactive = old_rdrs;
	ticket_rdrs_active = ticket_rdrs_inactive;
	splx(s);

	/* Purge the old rdr list */
	while ((rdr = TAILQ_FIRST(old_rdrs)) != NULL) {
		TAILQ_REMOVE(old_rdrs, rdr, entries);
		pf_dynaddr_remove(&rdr->saddr);
		pf_dynaddr_remove(&rdr->daddr);
		pf_dynaddr_remove(&rdr->raddr);
		uma_zfree(pf_rdr_pl, rdr);
	}
	return (0);
}

static int
pf_clrstates(void)
{
	struct pf_tree_node *n;
	int s;

	s = splsoftnet();
	RB_FOREACH(n, pf_state_tree, &tree_ext_gwy)
		n->state->expire = 0;
	pf_purge_expired_states();
	pf_status.states = 0;
	splx(s);
	return (0);
}


static int
shutdown_pf(void)
{
	struct pfioc_rule pr;
	struct pfioc_nat pn;
	struct pfioc_binat pbn;
	struct pfioc_rdr prd;
	int error = 0;

	mtx_lock(&pf_rule_mtx);
	untimeout(pf_purge_timeout, NULL, pf_expire_to);
	pf_status.running = 0;
	do {
		if ((error = pf_beginrules(&pr.ticket)))
			break;
		if ((error = pf_commitrules(&pr.ticket)))
			break;
		if ((error = pf_beginnats(&pn.ticket)))
			break;
		if ((error = pf_commitnats(&pn.ticket)))
			break;
		if ((error = pf_beginbinats(&pbn.ticket)))
			break;
		if ((error = pf_commitbinats(&pbn.ticket)))
			break;
		if ((error = pf_beginrdrs(&prd.ticket)))
			break;
		if ((error = pf_commitrdrs(&prd.ticket)))
			break;
		if ((error = pf_clrstates()))
			break;
	} while(0);
	mtx_unlock(&pf_rule_mtx);
	return (error);
}

/*
 * Need kernel PFIL_HOOKS option
 * XXX
 * It seems that it is not suitable for a mutiple kernel threads.
 */
static int
pf_check(void *ip, int hlen, struct ifnet *ifp, int dir, struct mbuf **m)
{
	int chk = pf_test(dir, ifp, m);
	if (chk != PF_PASS) {
		m_freem(*m);
	}
	return chk;
}

static int
pf_check6(void *ip, int hlen, struct ifnet *ifp, int dir, struct mbuf **m)
{
	int chk = pf_test6(dir, ifp, m);
	if (chk != PF_PASS) {
		m_freem(*m);
	}
	return chk;
}

static void
hook_pf(void)
{
	int s;

	s = splnet();
	/*
	 * XXX
	 * There is no easy way to get pfil header pointer with address
	 * family such as AF_INET, AF_INET6.
	 * Needs direct variable reference.
	 */
	pfil_add_hook(pf_check, PFIL_IN|PFIL_OUT,
		&inetsw[ip_protox[IPPROTO_IP]].pr_pfh);
#if defined(INET6)
	pfil_add_hook(pf_check6, PFIL_IN|PFIL_OUT,
		&inet6sw[ip_protox[IPPROTO_IPV6]].pr_pfh);
#endif
	splx(s);
}

static void
dehook_pf(void)
{
	int s;

	s = splnet();
	pfil_remove_hook(pf_check, PFIL_IN|PFIL_OUT,
		&inetsw[ip_protox[IPPROTO_IP]].pr_pfh);
#if defined(INET6)
	pfil_remove_hook(pf_check6, PFIL_IN|PFIL_OUT,
		&inet6sw[ip_protox[IPPROTO_IPV6]].pr_pfh);
#endif
	splx(s);
}

static int
pf_modevent(module_t mod, int type, void *data)
{
	int error = 0;

	switch(type) {
	case MOD_LOAD:
		init_pf_mutex();
		pf_dev = make_dev(&pf_cdevsw, 0, 0, 0, 0600, PF_NAME);
		init_zone_var();
		if (pfattach()) {
			destroy_dev(pf_dev);
			destroy_pf_mutex();
			error = ENOMEM;
		}
		else
			hook_pf();
		break;

	case MOD_UNLOAD:
		dehook_pf();
		shutdown_pf();
		cleanup_pf_zone();
		destroy_dev(pf_dev);
		destroy_pf_mutex();

		break;
	default:
		error = EINVAL;
		break;
	}
	return error;
}

static moduledata_t pf_mod = {
	"pf",
	pf_modevent,
	0
};

DECLARE_MODULE(pf, pf_mod, SI_SUB_PSEUDO, SI_ORDER_ANY);
MODULE_DEPEND(pf, pflog, PFLOG_MINVER, PFLOG_PREFVER, PFLOG_MAXVER);
MODULE_VERSION(pf, PF_MODVER);
#endif
